import os
import re
import sys
import warnings
from collections.abc import Iterable
from difflib import unified_diff
from functools import partial
from io import StringIO
from typing import IO, TYPE_CHECKING, Literal

from inifix._cli import Message, TaskResults, get_cpu_count
from inifix.io import _split_tokens, load

if TYPE_CHECKING:  # pragma: no cover
    from inifix._typing import AnyConfig

__all__ = ["format_string"]

PADDING_SIZE = 2


def _format_section(data: str) -> str:
    lines = data.splitlines()
    contents = []
    comments = []
    parameters = []
    values = []
    for line in lines:
        content, _, comment = line.partition("#")
        content = content.strip()
        if not content and not comment:
            continue
        contents.append(content)
        comments.append(comment.strip())
        if content.startswith("[") or not content:
            continue
        parameter, *value = _split_tokens(content)
        parameters.append(parameter)
        values.append(value)

    column_sizes: list[int] = []
    if not parameters:
        max_name_size = 0
    else:
        max_name_size = max(len(parameter) for parameter in parameters)
        num_value_cols = max(len(_) for _ in values)

        while len(column_sizes) < num_value_cols:
            new_column_idx = len(column_sizes)
            size = 0
            for v in values:
                if len(v) < new_column_idx + 1:
                    continue
                size = max(size, len(v[new_column_idx]))
            column_sizes.append(size + PADDING_SIZE)

    padded_name_col_size = max_name_size + 2 * PADDING_SIZE
    content_size = padded_name_col_size + sum(column_sizes) + PADDING_SIZE

    new_lines = []
    parameter_idx = 0
    for content, comment in zip(contents, comments, strict=True):
        new_line = ""
        offset = 0
        if content:
            if content.startswith("["):
                new_line += f"{content}\n"
            else:
                new_line += parameters[parameter_idx].ljust(padded_name_col_size)
                for val, size in zip(values[parameter_idx], column_sizes, strict=False):
                    new_line += val.ljust(size)
                parameter_idx += 1
                new_line = new_line.strip()
                offset = content_size - len(new_line)
        if comment:
            comm = f"# {comment}"
            new_line += comm.rjust(offset + len(comm))
        new_lines.append(new_line)
    res = "\n".join(new_lines)

    return res


def _finalize(res: str) -> str:
    # compress any duplicate new lines
    res = re.sub("\n+", "\n", res)
    # add one empty line before a new section
    res = re.sub(r"(.?)\n\[", r"\1\n\n[", res)
    # ensure there's exactly one newline at the EOF
    res = res.rstrip("\n")
    res += "\n"
    return res


def _iter_sections(fh: IO[str]) -> Iterable[str]:
    line = fh.readline()
    while line != "":
        content = [line]
        while (line := fh.readline()) != "" and not line.startswith("["):
            content.append(line)
        yield "".join(content)


def format_string(s: str, /) -> str:
    """Format a string representing the content of an inifile.

    This only operates on whitespace to and aims to maximize readability.
    Comments are preserved.

    Examples
    --------
    >>> import inifix
    >>> print(inifix.format_string(
    ...     "[Grid]\\n"
    ...     "X1-grid  1    0.0    1024          u    4.0\\n"
    ...     "X2-grid    1    0.0    256 u   1.0 # is ignored in 1D\\n"
    ...     "X3-grid      1    0.0    1       u    1.0  # is ignored in 1D and 2D"
    ... ))
    [Grid]
    X1-grid    1  0.0  1024  u  4.0
    X2-grid    1  0.0  256   u  1.0    # is ignored in 1D
    X3-grid    1  0.0  1     u  1.0    # is ignored in 1D and 2D
    <BLANKLINE>
    """
    fh = StringIO(s)
    content = []
    for s in _iter_sections(fh):
        content.append(_format_section(s))
    return _finalize("\n".join(content))


def iniformat(s: str, /) -> str:
    warnings.warn(
        "inifix.format.iniformat is deprecated since v5.0.0"
        "and will be removed in a future version. "
        "Please use inifix.format_string instead.",
        category=DeprecationWarning,
        stacklevel=2,
    )
    return format_string(s)


def main(argv: list[str] | None = None) -> int:
    from argparse import ArgumentParser
    from concurrent.futures import ThreadPoolExecutor

    parser = ArgumentParser()
    parser.add_argument("files", nargs="+")
    parser.add_argument(
        "--diff",
        action="store_true",
        help="Print the unified diff to stdout instead of editing files inplace",
    )
    parser.add_argument(
        "--report-noop",
        action="store_true",
        help="Explicitly log noops for files that are already formatted",
    )
    parser.add_argument(
        "--skip-validation",
        dest="validate",
        action="store_false",
        help="Skip validation step (formatting unvalidated data may lead to undefined behaviour)",
    )

    args = parser.parse_args(argv)
    closure = partial(
        _format_single_file_cli,
        args_validate=args.validate,
        args_report_noop=args.report_noop,
        args_diff=args.diff,
    )
    cpu_count = get_cpu_count()
    with ThreadPoolExecutor(max_workers=max(1, int(cpu_count / 2))) as executor:
        futures = [executor.submit(closure, file) for file in args.files]
        results = [f.result() for f in futures]

    for res in results:
        for message in res.messages:
            print(message.content, file=message.dest)

    return max(res.status for res in results)


def _format_single_file_cli(
    file: str, *, args_validate: bool, args_report_noop: bool, args_diff: bool
) -> TaskResults:
    status: Literal[0, 1] = 0
    messages: list[Message] = []

    if not os.path.isfile(file):
        status = 1
        messages.append(Message(f"Error: could not find {file}", sys.stderr))
        return TaskResults(status, messages)

    validate_baseline: AnyConfig = {}
    if args_validate:
        try:
            validate_baseline = load(file)
        except ValueError as exc:
            status = 1
            messages.append(Message(f"Error: {exc}", sys.stderr))
            return TaskResults(status, messages)

    with open(file, mode="rb") as fh:
        data = fh.read().decode("utf-8")
        # make sure newlines are always decoded as \n, even on windows
        data = data.replace("\r\n", "\n")

    fmted_data = format_string(data)

    if fmted_data == data:
        if args_report_noop:
            # printing to stderr so that we can pipe into cdiff in --diff mode
            messages.append(Message(f"{file} is already formatted", sys.stderr))
        return TaskResults(status, messages)

    if args_diff:
        diff = "\n".join(
            line.removesuffix("\n")
            for line in unified_diff(
                data.splitlines(), fmted_data.splitlines(), fromfile=file
            )
        )
        assert diff
        status = 1
        messages.append(Message(diff, sys.stdout))
    else:
        status = 1
        messages.append(Message(f"Fixing {file}", sys.stderr))
        if not os.access(file, os.W_OK):
            messages.append(
                Message(
                    f"Error: could not write to {file} (permission denied)", sys.stderr
                )
            )
            return TaskResults(status, messages)

        from tempfile import TemporaryDirectory

        with TemporaryDirectory(dir=os.path.dirname(file)) as tmpdir:
            tmpfile = os.path.join(tmpdir, "ini")
            with open(tmpfile, "wb") as bfh:
                bfh.write(fmted_data.encode("utf-8"))

            if args_validate and load(tmpfile) != validate_baseline:  # pragma: no cover
                messages.append(
                    Message(
                        f"Error: failed to format {file}: "
                        "formatted data compares unequal to unformatted data",
                        sys.stderr,
                    )
                )
                return TaskResults(status, messages)

            # this may still raise an error in the unlikely case of a race condition
            # (if permissions are changed between the look and the leap), but we
            # won't try to catch it unless it happens in production, because it is
            # difficult to test systematically.
            os.replace(tmpfile, file)

    return TaskResults(status, messages)


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



from .io import dump
from .io import dumps
from .io import load
from .io import loads
from .validation import validate_inifile_schema
from .format import format_string
from ._version import *

__all__ = [
    "dump",
    "dumps",
    "load",
    "loads",
    "validate_inifile_schema",
    "format_string",
    "__version__",
    "__version_tuple__",
]



from math import isnan
from typing import Any


def assert_item_equal(i1: Any, i2: Any) -> None:
    __tracebackhide__ = True
    assert type(i1) is type(i2)
    if type(i2) is float and isnan(i2):
        assert isnan(i1)
    else:
        assert i1 == i2


def assert_dict_equal(d1: dict, d2: dict) -> None:
    __tracebackhide__ = True
    # note that key insertion order matters in this comparison
    assert list(d2.keys()) == list(d1.keys())
    for v1, v2 in zip(d1.values(), d2.values(), strict=True):
        assert type(v1) is type(v2)
        if isinstance(v1, dict):
            assert_dict_equal(v1, v2)
        elif isinstance(v1, list):
            for i1, i2 in zip(v1, v2, strict=True):
                assert_item_equal(i1, i2)
        else:
            assert_item_equal(v1, v2)



# This file is automatically generated by Hatch
from typing import Final

__all__ = ["__version__", "__version_tuple__"]

__version__: Final = '5.1.2'
__version_tuple__: tuple[int, int, int] = (5, 1, 2)



import sys
from pathlib import Path

import pytest

DATA_DIR = Path(__file__).parent / "data"
INIFILES_PATHS = list(DATA_DIR.glob("*.ini")) + list(DATA_DIR.glob("*.cfg"))
INIFILES_IDS = [inifile.name[:-4] for inifile in INIFILES_PATHS]

INIFILES = dict(zip(INIFILES_PATHS, INIFILES_IDS, strict=True))

INIFILES_W_SECTIONS = {
    path: id
    for path, id in INIFILES.items()
    if "fargo" not in id and path.suffix != ".cfg"
}
INIFILES_WO_SECTIONS = {
    path: id for path, id in INIFILES.items() if path not in INIFILES_W_SECTIONS
}


@pytest.fixture()
def datadir():
    return DATA_DIR


@pytest.fixture(params=INIFILES.keys(), ids=INIFILES.values())
def inifile(request):
    return request.param


@pytest.fixture(params=INIFILES_W_SECTIONS.keys(), ids=INIFILES_W_SECTIONS.values())
def inifile_with_sections(request):
    return request.param


@pytest.fixture(params=INIFILES_WO_SECTIONS.keys(), ids=INIFILES_WO_SECTIONS.values())
def inifile_without_sections(request):
    return request.param


def pytest_report_header(config, start_path):
    if sys.version_info >= (3, 13):
        is_gil_enabled = sys._is_gil_enabled()
    else:
        is_gil_enabled = True

    return [
        f"{is_gil_enabled = }",
    ]



from __future__ import annotations

import os
import re
from collections.abc import Callable, Mapping
from functools import partial
from io import BufferedIOBase, IOBase
from typing import TYPE_CHECKING, Any, Literal, cast, overload

from inifix._more import always_iterable
from inifix._typing import (
    AnyConfig,
    Config_SectionsForbidden_ScalarsAllowed,
    Config_SectionsForbidden_ScalarsForbidden,
    Config_SectionsRequired_ScalarsAllowed,
    Config_SectionsRequired_ScalarsForbidden,
    Scalar,
    StrLike,
)
from inifix.enotation import ENotationIO
from inifix.validation import SCALAR_TYPES, validate_inifile_schema

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Iterable

    from _typeshed import GenericPath

    from inifix._typing import Config_SectionsAllowed_ScalarsForbidden

__all__ = [
    "dump",
    "dumps",
    "load",
    "loads",
]

SECTION_REGEXP = re.compile(r"\[(?P<title>[^(){}\[\]]+)\]\s*")


def _is_numeric(s: str) -> bool:
    try:
        float(s)
    except ValueError:
        return False
    else:
        return True


class Section(dict[str, Any]):
    def __init__(
        self,
        data: Mapping[str, Iterable[Scalar] | Scalar] | None = None,
        /,
        *,
        name: str | None = None,
    ) -> None:
        super().__init__()
        if data is not None:
            for k, v in data.items():
                self[k] = v
        self.name = name

    def __setitem__(self, k: Any, v: Any) -> None:
        if not isinstance(k, str):
            raise TypeError(f"Expected str keys. Received invalid key: {k}")

        if not (
            isinstance(v, SCALAR_TYPES)
            or (isinstance(v, list) and all(isinstance(_, SCALAR_TYPES) for _ in v))
        ):
            raise TypeError(
                "Expected all values to be scalars or lists of scalars. "
                f"Received invalid values {v}"
            )

        super().__setitem__(k, v)

    def _dump_to(self, storage: AnyConfig) -> None:
        if self.name is None:
            storage = cast(
                Config_SectionsForbidden_ScalarsAllowed
                | Config_SectionsForbidden_ScalarsForbidden,
                storage,
            )
            storage.update(self)
        else:
            storage = cast(
                Config_SectionsRequired_ScalarsAllowed
                | Config_SectionsRequired_ScalarsAllowed,
                storage,
            )
            storage[self.name] = dict(self)


# load helper functions


# this is more efficient than running str.partition in a loop
_SPLIT_COMMENTS = partial(str.split, sep="#", maxsplit=1)


def _normalize_data(data: StrLike) -> list[str]:
    # normalize text body `data` to parsable text lines
    if isinstance(data, bytes):
        data = data.decode("utf-8")
    return [line.strip() for (line, *_) in map(_SPLIT_COMMENTS, data.splitlines())]


_TOKEN = re.compile(r"""'[^']*'|"[^"]*"|\S+""")


def _split_tokens(data: str) -> list[str]:
    return _TOKEN.findall(data)


TRUTHY_STRINGS = frozenset({"true", "TRUE", "True", "yes", "YES", "Yes"})
FALSY_STRINGS = frozenset({"false", "FALSE", "False", "no", "NO", "No"})
ALL_BOOL_STRINGS = frozenset({*TRUTHY_STRINGS, *FALSY_STRINGS})

_RE_CASTERS: list[tuple[re.Pattern, Callable[[str], Any]]] = [
    (re.compile("(" + "|".join(TRUTHY_STRINGS) + ")"), lambda _: True),
    (re.compile("(" + "|".join(FALSY_STRINGS) + ")"), lambda _: False),
    (re.compile(r"^'.*'$"), lambda s: s[1:-1]),
    (re.compile(r'^".*"$'), lambda s: s[1:-1]),
]


def _auto_cast_aggressive(s: str) -> Scalar:
    try:
        f = float(s)
    except ValueError:
        pass
    else:
        if f.is_integer():
            return int(f)
        else:
            return f

    for regexp, caster in _RE_CASTERS:
        if regexp.fullmatch(s):
            return caster(s)

    return s


def _auto_cast_stable(s: str) -> Scalar:
    try:
        return int(s)
    except ValueError:
        pass

    try:
        return float(s)
    except ValueError:
        pass

    for regexp, caster in _RE_CASTERS:
        if regexp.fullmatch(s):
            return caster(s)

    return s


CasterFunction = Callable[[str], Scalar]


def _get_caster(integer_casting: Literal["stable", "aggressive"]) -> CasterFunction:
    match integer_casting:
        case "stable":
            return _auto_cast_stable
        case "aggressive":
            return _auto_cast_aggressive
        case _:
            raise ValueError(
                f"Unknown integer_casting value {integer_casting!r}. "
                "Expected 'stable' or 'aggressive'."
            )


def _tokenize_line(
    line: str,
    line_number: int,
    filename: str | None,
    caster: CasterFunction,
) -> tuple[str, list[Scalar]]:
    key, *raw_values = _split_tokens(line)
    if not raw_values:
        if filename is None:
            raise ValueError(f"Failed to parse line {line_number}: {line!r}")
        else:
            raise ValueError(f"Failed to parse {filename}:{line_number}:\n{line}")

    return key, [caster(v) for v in raw_values]


def _from_string(
    data: StrLike,
    *,
    parse_scalars_as_lists: bool,
    caster: CasterFunction,
    filename: str | None = None,
) -> AnyConfig:
    container: AnyConfig = {}
    lines = _normalize_data(data)
    section = Section()  # the default target is a nameless section
    for line_number, line in enumerate(lines, start=1):
        if not line:
            continue
        match = SECTION_REGEXP.fullmatch(line)
        if match is not None:
            if section:
                section._dump_to(container)
            section = Section(name=match["title"])
            continue

        values: Scalar | list[Scalar]
        key, values = _tokenize_line(
            line, filename=filename, line_number=line_number, caster=caster
        )
        if (not parse_scalars_as_lists) and len(values) == 1:
            values = values[0]
        section[key] = values
    section._dump_to(container)
    return container


def _from_file_descriptor(
    file: IOBase,
    *,
    parse_scalars_as_lists: bool,
    caster: CasterFunction,
) -> AnyConfig:
    filename = str(getattr(file, "name", repr(file)))
    data = file.read()
    lines = _normalize_data(data)
    if not "".join(lines):
        raise ValueError(f"{filename!r} appears to be empty.")
    return _from_string(
        data,
        filename=filename,
        parse_scalars_as_lists=parse_scalars_as_lists,
        caster=caster,
    )


def _from_path(
    file: GenericPath[str],
    *,
    parse_scalars_as_lists: bool,
    caster: CasterFunction,
) -> AnyConfig:
    file = os.fspath(file)
    with open(file, "rb") as fh:
        return _from_file_descriptor(
            fh,
            parse_scalars_as_lists=parse_scalars_as_lists,
            caster=caster,
        )


# dump helper functions


def _encode(v: Scalar) -> str:
    if isinstance(v, float):
        return ENotationIO.encode_preferential(v)
    elif isinstance(v, str) and (
        v == ""
        or re.search(r"\s", v) is not None
        or v in ALL_BOOL_STRINGS
        or _is_numeric(v)
    ):
        return repr(v)
    else:
        return str(v)


def _write(content: str, buffer: IOBase) -> None:
    if isinstance(buffer, BufferedIOBase):
        buffer.write(content.encode("utf-8"))
    else:
        buffer.write(content)


def _write_line(key: str, values: Iterable[Scalar] | Scalar, buffer: IOBase) -> None:
    val_repr = [_encode(v) for v in always_iterable(values)]
    _write(f"{key} {'  '.join(list(val_repr))}\n", buffer)


def _write_to_buffer(data: AnyConfig, buffer: IOBase) -> None:
    last = len(data) - 1
    for i, (key, val) in enumerate(data.items()):
        if not isinstance(val, dict):
            _write_line(key, val, buffer)
            continue
        _write(f"[{key}]\n", buffer)
        for k, v in val.items():
            _write_line(k, v, buffer)
        if i < last:
            _write("\n", buffer)


def _write_to_file(data: AnyConfig, file: GenericPath[str], /) -> None:
    if os.path.exists(file) and not os.access(file, os.W_OK):
        raise PermissionError(f"Cannot write to {file} (permission denied)")

    from tempfile import TemporaryDirectory

    with TemporaryDirectory(dir=os.path.dirname(file)) as tmpdir:
        tmpfile = os.path.join(tmpdir, "ini")
        with open(tmpfile, "wb") as fh:
            _write_to_buffer(data, fh)
        os.replace(tmpfile, file)


# narrowing return type on *two* keyword arguments at once:
# sections and parse_scalars_as_list

# these overloads are tricky to get right:
# - to avoid signature overlaps, we need to indicate default values in overloads
#   as often as possible
# - default values can only be specified where they match the annotated type
# - all arguments with default values need to come last in the signature
# - only keyword-only arguments can be re-arranged
# - since I'm overloading over two arguments at the same time, the above
#   *implies* that I can only do it for keyword-only arguments

# overloads are sorted from most to lease strict return type


@overload
def load(
    source: GenericPath[str] | IOBase,
    /,
    *,
    sections: Literal["forbid"],
    parse_scalars_as_lists: Literal[True],
    integer_casting: Literal["stable", "aggressive"] = "stable",
    skip_validation: bool = False,
) -> Config_SectionsForbidden_ScalarsForbidden: ...
@overload
def load(
    source: GenericPath[str] | IOBase,
    /,
    *,
    sections: Literal["require"],
    parse_scalars_as_lists: Literal[True],
    integer_casting: Literal["stable", "aggressive"] = "stable",
    skip_validation: bool = False,
) -> Config_SectionsRequired_ScalarsForbidden: ...
@overload
def load(
    source: GenericPath[str] | IOBase,
    /,
    *,
    sections: Literal["forbid"],
    parse_scalars_as_lists: Literal[False] = False,
    integer_casting: Literal["stable", "aggressive"] = "stable",
    skip_validation: bool = False,
) -> Config_SectionsForbidden_ScalarsAllowed: ...
@overload
def load(
    source: GenericPath[str] | IOBase,
    /,
    *,
    sections: Literal["require"],
    parse_scalars_as_lists: Literal[False] = False,
    integer_casting: Literal["stable", "aggressive"] = "stable",
    skip_validation: bool = False,
) -> Config_SectionsRequired_ScalarsAllowed: ...
@overload
def load(
    source: GenericPath[str] | IOBase,
    /,
    *,
    parse_scalars_as_lists: Literal[True],
    sections: Literal["allow"] = "allow",
    integer_casting: Literal["stable", "aggressive"] = "stable",
    skip_validation: bool = False,
) -> Config_SectionsAllowed_ScalarsForbidden: ...
@overload
def load(
    source: GenericPath[str] | IOBase,
    /,
    *,
    parse_scalars_as_lists: Literal[False] = False,
    sections: Literal["allow"] = "allow",
    integer_casting: Literal["stable", "aggressive"] = "stable",
    skip_validation: bool = False,
) -> AnyConfig: ...


def load(
    source: GenericPath[str] | IOBase,
    /,
    *,
    # parsing options
    parse_scalars_as_lists: bool = False,
    integer_casting: Literal["stable", "aggressive"] = "stable",
    # validation options
    sections: Literal["allow", "forbid", "require"] = "allow",
    skip_validation: bool = False,
) -> AnyConfig:
    """
    Parse data from a file.

    Parameters
    ----------
    source: any of the following
        - the name of a file to read from, (str, bytes or os.PathLike)
        - a readable handle. Both text and binary file modes are supported,
          though binary is preferred.
          In binary mode, we assume UTF-8 encoding.

    parse_scalars_as_lists: bool (default: False)
        if set to True, all values will be parses as lists of scalars,
        even for parameters comprised of a single scalar.
        This also has the effect of narrowing down the return type as seen
        by static type checkers.

        .. versionadded: 3.1.0

    integer_casting: 'stable' (default) or 'aggressive'
        casting strategy for numbers written in decimal notations, such as '1.',
        '2.0' or '3e0'. By default, perform roundtrip-stable casting (i.e., cast
        as Python floats). Setting `integer_casting='aggressive'` will instead
        parse these as Python ints, matching the behavior of inifix 4.5

        .. versionadded: 5.0.0

    sections: 'allow' (default), 'forbid' or 'require'
        use sections='forbid' to invalidate any section found,
        or sections='require' to invalidate a sectionless structure.
        Default mode (sections='allow') permits both.
        sections='forbid' and sections='require' both have the effect of
        narrowing down the return type as seen by static type checkers.
        This parameter has no effect at runtime when combined with
        skip_validation=True.

        .. versionadded: 5.1.0

    skip_validation: bool (default: False)
        if set to True, input is not validated. This can be used to speedup
        io operations trusted input or to work around bugs with the validation
        routine. Use with caution.
        This parameter intentionally has no effect on static type checking.

        .. versionadded: 4.1.0

    See Also
    --------
    inifix.loads
    """
    caster = _get_caster(integer_casting)

    if isinstance(source, IOBase):
        config = _from_file_descriptor(
            source,
            parse_scalars_as_lists=parse_scalars_as_lists,
            caster=caster,
        )
    else:
        assert isinstance(source, str | bytes | os.PathLike)
        config = _from_path(
            source,
            parse_scalars_as_lists=parse_scalars_as_lists,
            caster=caster,
        )

    if not skip_validation:
        validate_inifile_schema(config, sections=sections)
    return config


@overload
def loads(
    source: str,
    /,
    *,
    sections: Literal["forbid"],
    parse_scalars_as_lists: Literal[True],
    integer_casting: Literal["stable", "aggressive"] = "stable",
    skip_validation: bool = False,
) -> Config_SectionsForbidden_ScalarsForbidden: ...
@overload
def loads(
    source: str,
    /,
    *,
    sections: Literal["require"],
    parse_scalars_as_lists: Literal[True],
    integer_casting: Literal["stable", "aggressive"] = "stable",
    skip_validation: bool = False,
) -> Config_SectionsRequired_ScalarsForbidden: ...
@overload
def loads(
    source: str,
    /,
    *,
    sections: Literal["forbid"],
    parse_scalars_as_lists: Literal[False] = False,
    integer_casting: Literal["stable", "aggressive"] = "stable",
    skip_validation: bool = False,
) -> Config_SectionsForbidden_ScalarsAllowed: ...
@overload
def loads(
    source: str,
    /,
    *,
    sections: Literal["require"],
    parse_scalars_as_lists: Literal[False] = False,
    integer_casting: Literal["stable", "aggressive"] = "stable",
    skip_validation: bool = False,
) -> Config_SectionsRequired_ScalarsAllowed: ...
@overload
def loads(
    source: str,
    /,
    *,
    parse_scalars_as_lists: Literal[True],
    sections: Literal["allow"] = "allow",
    integer_casting: Literal["stable", "aggressive"] = "stable",
    skip_validation: bool = False,
) -> Config_SectionsAllowed_ScalarsForbidden: ...
@overload
def loads(
    source: str,
    /,
    *,
    parse_scalars_as_lists: Literal[False] = False,
    sections: Literal["allow"] = "allow",
    integer_casting: Literal["stable", "aggressive"] = "stable",
    skip_validation: bool = False,
) -> AnyConfig: ...


def loads(
    source: str,
    /,
    *,
    # parsing options
    parse_scalars_as_lists: bool = False,
    integer_casting: Literal["stable", "aggressive"] = "stable",
    # validation options
    sections: Literal["allow", "forbid", "require"] = "allow",
    skip_validation: bool = False,
) -> AnyConfig:
    """
    Parse data from a string.

    Parameters
    ----------
    source: str
        The content of a parameter file (has to be inifix format-compliant)

    parse_scalars_as_lists: bool (default: False)
        if set to True, all values will be parses as lists of scalars,
        even for parameters comprised of a single scalar.
        This also has the effect of narrowing down the return type as seen
        by static type checkers.

        .. versionadded: 3.1.0

    integer_casting: 'stable' (default) or 'aggressive'
        casting strategy for numbers written in decimal notations, such as '1.',
        '2.0' or '3e0'. By default, perform roundtrip-stable casting (i.e., cast
        as Python floats). Setting `integer_casting='aggressive'` will instead
        parse these as Python ints, matching the behavior of inifix 4.5

        .. versionadded: 5.0.0

    sections: 'allow' (default), 'forbid' or 'require'
        use sections='forbid' to invalidate any section found,
        or sections='require' to invalidate a sectionless structure.
        Default mode (sections='allow') permits both.
        sections='forbid' and sections='require' both have the effect of
        narrowing down the return type as seen by static type checkers.
        This parameter has no effect at runtime when combined with
        skip_validation=True.

        .. versionadded: 5.1.0

    skip_validation: bool (default: False)
        if set to True, input is not validated. This can be used to speedup
        io operations trusted input or to work around bugs with the validation
        routine. Use with caution.
        This parameter intentionally has no effect on static type checking.

        .. versionadded: 4.1.0

    See Also
    --------
    inifix.load
    """
    caster = _get_caster(integer_casting)
    retv = _from_string(
        source, parse_scalars_as_lists=parse_scalars_as_lists, caster=caster
    )

    if not skip_validation:
        validate_inifile_schema(retv, sections=sections)
    return retv


def dump(
    data: AnyConfig,
    /,
    file: GenericPath[str] | IOBase,
    *,
    # validation options
    sections: Literal["allow", "forbid", "require"] = "allow",
    skip_validation: bool = False,
) -> None:
    """
    Write data to a file.

    Parameters
    ----------
    data: dict
        has to be inifix format-compliant

    file: any of the following
        - the name of a file to write to (str, bytes or os.PathLike)
        - a writable handle. Both text and binary file modes are supported,
          though binary is preferred.
          In binary mode, data is encoded as UTF-8.

    sections: 'allow' (default), 'forbid' or 'require'
        use sections='forbid' to invalidate any section found,
        or sections='require' to invalidate a sectionless structure.
        Default mode (sections='allow') permits both.
        sections='forbid' and sections='require' both have the effect of
        narrowing down the return type as seen by static type checkers.
        This parameter has no effect at runtime when combined with
        skip_validation=True.

        .. versionadded: 5.1.0

    skip_validation: bool (default: False)
        if set to True, input is not validated. This can be used to speedup
        io operations trusted input or to work around bugs with the validation
        routine. Use with caution.
        This parameter intentionally has no effect on static type checking.

        .. versionadded: 4.1.0

    See Also
    --------
    inifix.dumps
    """
    if not skip_validation:
        validate_inifile_schema(data, sections=sections)

    if isinstance(file, IOBase):
        _write_to_buffer(data, file)
    else:
        _write_to_file(data, file)


def dumps(
    data: AnyConfig,
    /,
    *,
    # validation options
    sections: Literal["allow", "forbid", "require"] = "allow",
    skip_validation: bool = False,
) -> str:
    """
    Convert data to a string.

    Parameters
    ----------
    data: dict
        has to be inifix format-compliant

    sections: 'allow' (default), 'forbid' or 'require'
        use sections='forbid' to invalidate any section found,
        or sections='require' to invalidate a sectionless structure.
        Default mode (sections='allow') permits both.
        sections='forbid' and sections='require' both have the effect of
        narrowing down the return type as seen by static type checkers.
        This parameter has no effect at runtime when combined with
        skip_validation=True.

        .. versionadded: 5.1.0

    skip_validation: bool (default: False)
        if set to True, input is not validated. This can be used to speedup
        io operations trusted input or to work around bugs with the validation
        routine. Use with caution.
        This parameter intentionally has no effect on static type checking.

        .. versionadded: 4.1.0

    Returns
    -------
    string representing the content of a parameter file

    See Also
    --------
    inifix.dump
    """
    from io import BytesIO

    s = BytesIO()
    dump(data, file=s, skip_validation=skip_validation, sections=sections)
    return s.getvalue().decode("utf-8")



import os
import sys
from typing import Literal, NamedTuple, TextIO


class Message(NamedTuple):
    content: str
    dest: TextIO


class TaskResults(NamedTuple):
    status: Literal[0, 1]
    messages: list[Message]


def get_cpu_count() -> int:
    # this function exists primarily to be mocked
    # instead of something we don't own
    base_cpu_count: int | None
    if sys.version_info >= (3, 13):
        base_cpu_count = os.process_cpu_count()
    elif hasattr(os, "sched_getaffinity"):
        # this function isn't available on all platforms
        base_cpu_count = len(os.sched_getaffinity(0))
    else:  # pragma: no cover
        # this proxy is good enough in most situations
        base_cpu_count = os.cpu_count()
    return base_cpu_count or 1



import argparse
import os
import sys

from inifix.io import load


def main(argv: list[str] | None = None) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("files", nargs="+")

    args = parser.parse_args(argv)
    retv = 0
    for file in args.files:
        if not os.path.isfile(file):
            print(f"Error: could not find {file}", file=sys.stderr)
            retv = 1
            continue
        try:
            load(file)
        except ValueError as exc:
            print(f"Failed to validate {file}:\n  {exc}", file=sys.stderr)
            retv = 1
        else:
            print(f"Validated {file}")

    return retv


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



import shutil
from pathlib import Path

import pytest

from inifix.format import main as inifix_format_cli
from inifix.validate import main as inifix_validate_cli

DATA_DIR = Path(__file__).parent / "data"


N_FILES = 257


@pytest.fixture
def unformatted_files(tmp_path):
    in_file = DATA_DIR / "format-in.ini"
    files = []
    for file_no in range(N_FILES):
        new_file = tmp_path / f"{file_no}.ini"
        shutil.copy(in_file, new_file)
        files.append(new_file.resolve())

    yield files


def test_inifix_format_cli(unformatted_files, capsys):
    ret = inifix_format_cli([str(f) for f in unformatted_files])
    assert ret != 0

    out, err = capsys.readouterr()
    assert out == ""

    # order of lines doesn't matter and is not guaranteed
    err_lines = err.splitlines()
    assert set(err_lines) == {f"Fixing {file}" for file in unformatted_files}

    expected = (DATA_DIR / "format-out.ini").read_text()
    for file in unformatted_files:
        body = file.read_text()
        assert body == expected


def test_inifix_validate_cli(unformatted_files, capsys):
    ret = inifix_validate_cli([str(f) for f in unformatted_files])
    assert ret == 0

    out, err = capsys.readouterr()
    assert err == ""

    # order of lines doesn't matter and is not guaranteed
    out_lines = out.splitlines()
    assert set(out_lines) == {f"Validated {file}" for file in unformatted_files}



from inifix.io import load


def test_fargo_dummy(datadir):
    conf = load(datadir / "fargo-dummy.ini")
    expected = {
        "x": 1,
        "y": 2,
        "z": 3,
    }
    assert conf == expected



import re
import sys
from enum import Enum, auto
from typing import Any, Literal

from inifix._more import always_iterable
from inifix._typing import AnyConfig

if sys.version_info >= (3, 11):
    from typing import assert_never
else:
    from typing_extensions import assert_never

_PARAM_NAME_REGEXP = re.compile(r"[-\.\w]+")
SCALAR_TYPES = (int, float, bool, str)


class SectionsMode(Enum):
    ALLOW = auto()
    FORBID = auto()
    REQUIRE = auto()


def _uses_invalid_chars(s: str) -> bool:
    ma = re.fullmatch(_PARAM_NAME_REGEXP, s)
    return ma is None


def validate_elementary_item(key: Any, value: Any) -> None:
    if not isinstance(key, str):
        raise ValueError(
            f"Invalid schema: received key '{key}' with type '{type(key)}', "
            "expected a str"
        )
    if len(key) == 0:
        raise ValueError("Invalid schema: received an empty str as key")
    if _uses_invalid_chars(key):
        raise ValueError(
            f"Invalid schema: received key {key!r}, "
            "expected only alphanumeric characters or special characters from {'-', '.'}"
        )
    if not key[0].isalpha():
        raise ValueError(
            f"Invalid schema: received key {key!r}, "
            "keys are expected to start with a letter"
        )
    if not isinstance(value, (*SCALAR_TYPES, list)):
        raise ValueError(
            f"Invalid schema: received value with type '{type(value)}', "
            "expected an int, float, bool, str, or list"
        )
    for ev in always_iterable(value):
        if not isinstance(ev, SCALAR_TYPES):
            raise ValueError(
                f"Invalid schema: received value '{ev}' with type '{type(ev)}', "
                "expected an int, float, bool or str"
            )


def validate_inifile_schema(
    data: AnyConfig,
    /,
    *,
    sections: Literal["allow", "forbid", "require"] = "allow",
) -> None:
    """
    Validate structure of a dictionary as an inifix-compliant configuration.

    Parameters
    ----------
    data: dict
      the candidate configuration to be (in)validated.

    sections: 'allow' (default), 'forbid' or 'require'
      use sections='forbid' to invalidate any section found,
      or sections='require' to invalidate a sectionless structure.
      Default mode (sections='allow') permits both.

      .. versionadded: 5.1.0

    Raises
    ------
    TypeError: for unrecognized values in parameter sections.
    ValueError: if and only if data does not conform to the expected schema.

    """
    match sections:
        case "allow":
            sections_mode = SectionsMode.ALLOW
        case "forbid":
            sections_mode = SectionsMode.FORBID
        case "require":
            sections_mode = SectionsMode.REQUIRE
        case _:
            raise TypeError(
                "Unknown value for parameter sections. "
                f"Got {sections=!r}, expected 'allow', 'forbid' or 'require'"
            )

    for k, v in data.items():
        if not isinstance(k, str):
            raise ValueError(
                f"Invalid schema: received key '{k}' "
                f"with type '{type(k)}', expected a str"
            )

        if isinstance(v, dict):
            match sections_mode:
                case SectionsMode.ALLOW | SectionsMode.REQUIRE:
                    for kk, vv in v.items():
                        validate_elementary_item(kk, vv)
                case SectionsMode.FORBID:
                    raise ValueError(
                        "Invalid schema: sections were explicitly forbidden, "
                        f"but one was found under key '{k}'"
                    )
                case _:  # pragma: no cover
                    assert_never(sections_mode)

        else:
            match sections_mode:
                case SectionsMode.ALLOW | SectionsMode.FORBID:
                    validate_elementary_item(k, v)
                case SectionsMode.REQUIRE:
                    raise ValueError(
                        "Invalid schema: sections were explicitly required, "
                        "but the following key/value pair was found outside of "
                        f"any section: '{k}', {v}"
                    )
                case _:  # pragma: no cover
                    assert_never(sections_mode)



import difflib
import os
import re
import tempfile
from io import BytesIO
from pathlib import Path
from stat import S_IREAD

import pytest
from hypothesis import example, given
from hypothesis import strategies as st

from inifix.io import (
    ALL_BOOL_STRINGS,
    FALSY_STRINGS,
    TRUTHY_STRINGS,
    Section,
    _auto_cast_stable,
    _tokenize_line,
    dump,
    dumps,
    load,
    loads,
)

from .utils import assert_dict_equal


@pytest.mark.parametrize(
    "invalid_data", ["", "SingleToken", "[InvalidSectionName", "InvalidSection]"]
)
def test_tokenizer(invalid_data):
    with pytest.raises(ValueError):
        _tokenize_line(
            invalid_data,
            filename="fake_filename",
            line_number=-1,
            caster=_auto_cast_stable,
        )


def test_unit_read(inifile):
    load(inifile)


def test_idempotent_io(inifile):
    data0 = load(inifile)
    with tempfile.TemporaryDirectory() as tmpdir:
        save1 = Path(tmpdir) / "save1"
        save2 = Path(tmpdir) / "save2"
        dump(data0, save1)
        data1 = load(save1)
        dump(data1, save2)

        text1 = save1.read_text().split("\n")
        text2 = save2.read_text().split("\n")

        diff = "".join(difflib.context_diff(text1, text2))
        assert not diff


@pytest.mark.parametrize(
    "data, expected",
    [
        ("""val content\n""", {"val": "content"}),
        ("""val 'content'\n""", {"val": "content"}),
        ("""val content"\n""", {"val": 'content"'}),
        ("""val "content\n""", {"val": '"content'}),
        ("""val content'\n""", {"val": "content'"}),
        ("""val 'content\n""", {"val": "'content"}),
        ("""val "content"\n""", {"val": "content"}),
        ("""val "content'"\n""", {"val": "content'"}),
        ("""val '"content"'\n""", {"val": '"content"'}),
        ("""val "'content'"\n""", {"val": "'content'"}),
        ("""val true\n""", {"val": True}),
        ("""val "true"\n""", {"val": "true"}),
        ("""val 'true'\n""", {"val": "true"}),
        ("""val false\n""", {"val": False}),
        ("""val "false"\n""", {"val": "false"}),
        ("""val 'false'\n""", {"val": "false"}),
        ("""val 1\n""", {"val": 1}),
        ("""val "1"\n""", {"val": "1"}),
        ("""val '1'\n""", {"val": "1"}),
        ("""val 1e2\n""", {"val": 100}),
        ("""val "1e2"\n""", {"val": "1e2"}),
        ("""val '1e2'\n""", {"val": "1e2"}),
        (
            "name true 'true'     'steven bacon'  1",
            {"name": [True, "true", "steven bacon", 1]},
        ),
        (
            # see https://github.com/neutrinoceros/inifix/issues/167
            "dogs Idefix.000000",
            {"dogs": "Idefix.000000"},
        ),
    ],
)
def test_string_casting(data, expected):
    mapping = loads(data)
    assert mapping == expected


@pytest.mark.parametrize(
    "data, expected",
    [
        pytest.param(
            '''[Spam]\nEggs "Bacon  Saussage"''',
            {"Spam": {"Eggs": "Bacon  Saussage"}},
            id="simple strings",
        ),
        pytest.param(
            '''numbers 1 '1' 2.2 '2.2' 3.3.3 "3.3.3"''',
            {"numbers": [1, "1", 2.2, "2.2", "3.3.3", "3.3.3"]},
            id="numbers",
        ),
        pytest.param(
            '''vals  1 'bacon' 2.2 "saussage" True no "a b c d"''',
            {"vals": [1, "bacon", 2.2, "saussage", True, False, "a b c d"]},
            id="mixed bag",
        ),
    ],
)
def test_idempotent_string_parsing(data, expected):
    initial_mapping = loads(data)
    assert initial_mapping == expected
    round_str = dumps(initial_mapping)
    round_mapping = loads(round_str)
    assert round_mapping == initial_mapping


@pytest.mark.parametrize("s", ALL_BOOL_STRINGS)
def test_bool_strings(s):
    if s in TRUTHY_STRINGS:
        b = True
    elif s in FALSY_STRINGS:
        b = False
    else:  # pragma: no cover
        raise RuntimeError
    # test unsupported case
    S = s.title().swapcase()
    data = loads(f"a {s} '{s}' {S}")
    assert data == {"a": [b, s, S]}


def test_section_init():
    data = {
        "dummy": [0.0001, True],
        "thisparameternameshouldprobablybeshorter": [45, 68],
        "thisoneisshorter": [15, 68, 774, 6, 7, 5],
        "faultyTowers": 42,
    }
    s1 = Section(data)
    assert s1.name is None

    s2 = Section(data, name="test")
    assert s2.name == "test"


def test_invalid_section_value():
    val = frozenset((1, 2, 3))
    data = {
        "yes": val,
    }
    with pytest.raises(
        TypeError,
        match=re.escape(
            "Expected all values to be scalars or lists of scalars. "
            f"Received invalid values {val}"
        ),
    ):
        Section(data)


def test_invalid_section_key():
    data = {
        1: True,
    }
    with pytest.raises(
        TypeError, match=re.escape("Expected str keys. Received invalid key: 1")
    ):
        Section(data)


@pytest.mark.parametrize("mode", ("w", "wb"))
def test_dump_to_file_descriptor(mode, inifile, tmp_path):
    conf = load(inifile)

    file = tmp_path / "save.ini"
    with open(file, mode=mode) as fh:
        dump(conf, fh)

    # a little weak but better than just testing that the file isn't empty
    new_body = file.read_text()
    for key, val in conf.items():
        if isinstance(val, dict):
            assert f"[{key}]\n" in new_body


def test_dump_to_file_path(inifile, tmp_path):
    conf = load(inifile)

    # pathlib.Path obj
    file1 = tmp_path / "save1.ini"
    dump(conf, file1, skip_validation=True)
    body1 = file1.read_text()

    # str
    file2 = tmp_path / "save2.ini"
    sfile2 = str(file2)
    dump(conf, sfile2, skip_validation=True)
    body2 = file2.read_text()

    assert body1 == body2
    for key, val in conf.items():
        if isinstance(val, dict):
            assert f"[{key}]\n" in body2


def test_load_empty_file(tmp_path):
    target = tmp_path / "empty_file"
    target.touch()
    with pytest.raises(
        ValueError, match=re.escape(f"{str(target)!r} appears to be empty.")
    ):
        load(target)


@pytest.mark.parametrize("mode", ("r", "rb"))
def test_load_from_descriptor(inifile, mode):
    with open(inifile, mode=mode) as fh:
        load(fh)


def test_loads_empty_str():
    ret = loads("")
    assert ret == {}


def test_loads_invalid_str():
    with pytest.raises(ValueError, match="Failed to parse line 1: 'invalid'"):
        loads("invalid")


def test_loads_dumps_roundtrip(inifile):
    with open(inifile) as fh:
        data = fh.read()
    d1 = loads(data)
    s1 = dumps(d1)
    d2 = loads(s1)
    assert d1 == d2


def test_error_read_only_file(tmp_path):
    target = tmp_path / "ini"
    target.touch()
    os.chmod(target, S_IREAD)

    data = {"names": ["bob", "alice"]}

    with pytest.raises(
        PermissionError,
        match=re.escape(f"Cannot write to {target} (permission denied)"),
    ):
        dump(data, target)


def test_read_from_binary_io():
    b = BytesIO(b"var 1 2 a b 'hello world'")
    data = load(b)
    assert data == {"var": [1, 2, "a", "b", "hello world"]}


def test_parse_scalars_as_lists(inifile):
    def _validate(conf):
        for value in conf.values():
            if isinstance(value, dict):
                _validate(value)
            else:
                assert type(value) is list  # noqa: E721

    conf1 = load(inifile, parse_scalars_as_lists=True)
    with open(inifile) as fh:
        conf2 = load(fh, parse_scalars_as_lists=True)

    _validate(conf1)
    _validate(conf2)


def test_skip_validation(monkeypatch, tmp_path):
    import inifix

    def _mp_validate_inifile_schema(d, /, **kwargs):
        raise ValueError("gotcha")

    # cannot monkeypatch inifix.validation.validate_inifile_schema directly ...
    monkeypatch.setattr(
        inifix.io, "validate_inifile_schema", _mp_validate_inifile_schema
    )
    ctx = pytest.raises(ValueError, match="gotcha")

    data = "[Static Grid Output]\n" "dbl.h5    -1.0  -1"

    with ctx:
        loads(data)

    with open(tmp_path / "data.ini", "w") as fh:
        fh.write(data)

    with ctx:
        load(tmp_path / "data.ini")

    conf1 = loads(data, skip_validation=True)
    conf2 = load(tmp_path / "data.ini", skip_validation=True)
    assert conf1 == conf2


@pytest.mark.parametrize(
    "kwargs",
    [
        pytest.param({}, id="implicitly stable"),
        pytest.param({"integer_casting": "stable"}, id="explicitly stable"),
    ],
)
@given(
    st.lists(
        st.one_of(
            st.integers(),
            st.floats(),
            st.booleans(),
            st.text(
                alphabet=st.characters(
                    codec="utf-8",
                    categories=("Nd", "L"),
                    include_characters=[" "],
                )
            ),
        )
    ).filter(lambda L: len(L) > 0)
)
@example([1, ""])  # regression test for gh-243
def test_roundtrip_stability_generated(kwargs, L):
    if len(L) == 1:
        kwargs.setdefault("parse_scalars_as_lists", True)
    data1 = {"a": L}
    data1_rt = loads(dumps(data1), **kwargs)
    assert_dict_equal(data1_rt, data1)

    data2 = {"Section 1": data1, "Section 2": data1}
    data2_rt = loads(dumps(data2), **kwargs)
    assert_dict_equal(data2_rt, data2)


def test_aggressive_integer_casting():
    input_data = "opt 0 1. 2.0 3e0 4.5"
    data = loads(input_data, integer_casting="aggressive")

    expected = {"opt": [0, 1, 2, 3, 4.5]}
    assert list(data.keys()) == ["opt"]
    for item, expected_item in zip(data["opt"], expected["opt"], strict=True):
        assert item == expected_item
        assert type(item) is type(expected_item)


def test_unknown_integer_casting():
    input_data = "opt 0 1. 2.0 3e0 4.5"
    with pytest.raises(
        ValueError,
        match="Unknown integer_casting value 'unknown_strategy'.",
    ):
        loads(input_data, integer_casting="unknown_strategy")



from math import isnan

import pytest
from hypothesis import given
from hypothesis import strategies as st

import inifix
from inifix.enotation import ENotationIO
from inifix.io import _auto_cast_aggressive, _auto_cast_stable

BASE_BOOLS = [
    ("True", True),
    ("true", True),
    ("Yes", True),
    ("yes", True),
    ("False", False),
    ("false", False),
    ("No", False),
    ("no", False),
]


@pytest.mark.parametrize("caster", [_auto_cast_stable, _auto_cast_aggressive])
@pytest.mark.parametrize("s, expected", BASE_BOOLS)
def test_bool_cast(caster, s, expected):
    assert caster(s) is expected


@pytest.mark.parametrize("caster", [_auto_cast_stable, _auto_cast_aggressive])
@pytest.mark.parametrize("s", ["tdsk", "1213", "Treu", "Flsae", "flkj"])
def test_bool_cast_invalid(caster, s):
    assert type(caster(s)) is not bool  # noqa: E721


@pytest.mark.parametrize("s, expected", BASE_BOOLS)
def test_bool_cast_integration(s, expected, tmp_path):
    with open(tmp_path / "dummy.ini", "w") as fh:
        fh.write(f"dummy {s}")
    d = inifix.load(tmp_path / "dummy.ini")
    assert d == {"dummy": expected}


@given(st.integers())
def test_unambiguous_int_cast_stable(i):
    s = str(i)
    res = _auto_cast_stable(s)
    assert res == i
    assert type(res) is int


# aggressively casting to float first leads to loss of precision beyond a threshold
@given(st.integers(min_value=-9_007_199_254_740_992, max_value=9_007_199_254_740_992))
def test_unambiguous_int_cast_aggressive(i):
    s = str(i)
    res = _auto_cast_aggressive(s)
    assert res == i
    assert type(res) is int


@pytest.mark.parametrize(
    "caster, expected_type",
    [
        (_auto_cast_stable, float),
        (_auto_cast_aggressive, int),
    ],
)
@given(st.integers())
def test_int_like_floats_casting(caster, expected_type, i):
    f = float(i)
    for s in (str(f), ENotationIO.encode(i)):
        res = caster(s)
        assert res == f
        assert type(res) is expected_type
        if expected_type is float:
            assert res.is_integer()


@pytest.mark.parametrize("caster", [_auto_cast_stable, _auto_cast_aggressive])
@given(st.floats())
def test_unambiguous_float_casting(caster, f):
    s = str(f)
    res = caster(s)
    if isnan(f):
        assert isnan(res)
    else:
        assert res == float(s)
    assert isinstance(res, int | float)
    if caster is _auto_cast_aggressive and f.is_integer():
        assert type(res) is int
    else:
        assert type(res) is float



from inifix.io import load


def test_pluto_disk_planet(datadir):
    conf = load(datadir / "pluto-DiskPlanet.ini")
    expected = [
        "Grid",
        "Chombo Refinement",
        "Time",
        "Solver",
        "Boundary",
        "Static Grid Output",
        "Chombo HDF5 output",
        "Parameters",
    ]
    sections = list(conf.keys())
    assert sections == expected



import pytest

from inifix.io import dump, dumps, load, loads
from inifix.validation import validate_inifile_schema

from .utils import assert_dict_equal


def test_validate_known_files(inifile):
    conf = load(inifile)
    # implicit form
    validate_inifile_schema(conf)

    # explicit form
    validate_inifile_schema(conf, sections="allow")


def test_validate_known_files_without_sections(inifile_without_sections, tmp_path):
    conf1 = load(inifile_without_sections)
    validate_inifile_schema(conf1, sections="forbid")

    expected_msg = (
        "Invalid schema: sections were explicitly required, "
        "but the following key/value pair was found outside of "
        "any section"
    )
    with pytest.raises(ValueError, match=expected_msg):
        validate_inifile_schema(conf1, sections="require")

    with pytest.raises(ValueError, match=expected_msg):
        load(inifile_without_sections, sections="require")

    # check that sections=... doesn't have any effect when combined with skip_validation=True
    conf2 = load(inifile_without_sections, sections="require", skip_validation=True)
    assert_dict_equal(conf2, conf1)

    save_file = tmp_path / "test"
    with pytest.raises(ValueError, match=expected_msg):
        dump(conf1, save_file, sections="require")
    assert not save_file.exists()

    dump(conf1, save_file, sections="require", skip_validation=True)
    assert save_file.is_file()

    with pytest.raises(ValueError, match=expected_msg):
        dumps(conf1, sections="require")

    s = dumps(conf1, sections="require", skip_validation=True)

    conf3 = loads(s)
    assert_dict_equal(conf3, conf1)
    with pytest.raises(ValueError, match=expected_msg):
        loads(s, sections="require")

    conf4 = loads(s, sections="require", skip_validation=True)
    assert_dict_equal(conf4, conf3)


def test_validate_known_files_with_sections(inifile_with_sections, tmp_path):
    conf1 = load(inifile_with_sections)
    validate_inifile_schema(conf1, sections="require")
    expected_msg = (
        "Invalid schema: sections were explicitly forbidden, "
        "but one was found under key"
    )
    with pytest.raises(ValueError, match=expected_msg):
        validate_inifile_schema(conf1, sections="forbid")

    with pytest.raises(ValueError, match=expected_msg):
        load(inifile_with_sections, sections="forbid")

    # check that sections=... doesn't have any effect when combined with skip_validation=True
    conf2 = load(inifile_with_sections, sections="forbid", skip_validation=True)
    assert_dict_equal(conf2, conf1)

    save_file = tmp_path / "test"
    with pytest.raises(ValueError, match=expected_msg):
        dump(conf1, save_file, sections="forbid")
    assert not save_file.exists()

    dump(conf1, save_file, sections="forbid", skip_validation=True)
    assert save_file.is_file()

    with pytest.raises(ValueError, match=expected_msg):
        dumps(conf1, sections="forbid")

    s = dumps(conf1, sections="forbid", skip_validation=True)

    conf3 = loads(s)
    assert_dict_equal(conf3, conf1)
    with pytest.raises(ValueError, match=expected_msg):
        loads(s, sections="forbid")

    conf4 = loads(s, sections="forbid", skip_validation=True)
    assert_dict_equal(conf4, conf3)


@pytest.mark.parametrize(
    "invalid_conf",
    [
        pytest.param({1: {"param": 1}}, id="section_no_str"),
        pytest.param({"section": {"": 1}}, id="empty_param_name"),
        pytest.param({"section": {"§": 1}}, id="invalid_param_name-1"),
        pytest.param({"section": {"a§": 1}}, id="invalid_param_name-2"),
        pytest.param({"section": {1: "one"}}, id="param_name_no_str"),
        pytest.param({"section": [[1, 2], [3, 4]]}, id="non_flat_section"),
        pytest.param(
            {"section": {"param": [[1, 2], [3, 4]]}}, id="nested_non_flat_sequence"
        ),
        pytest.param(
            {"section": {"subsection": {"subsubsection": 1}}}, id="too_deep_nesting"
        ),
    ],
)
def test_dump_invalid_conf(invalid_conf, tmp_path):
    with pytest.raises(ValueError, match=r"^(Invalid schema)"):
        dump(invalid_conf, tmp_path / "save.ini")


def test_unknown_sections_value():
    with pytest.raises(TypeError):
        validate_inifile_schema({}, sections="unknown-value")



from typing import TypeAlias

# not quite typing.AnyStr : this is not a constrained type variable
StrLike: TypeAlias = str | bytes
Scalar: TypeAlias = int | float | bool | str

Section_ScalarsAllowed: TypeAlias = dict[str, list[Scalar] | Scalar]
Section_ScalarsForbidden: TypeAlias = dict[str, list[Scalar]]

Config_SectionsRequired_ScalarsForbidden: TypeAlias = dict[
    str, Section_ScalarsForbidden
]
Config_SectionsRequired_ScalarsAllowed: TypeAlias = dict[str, Section_ScalarsAllowed]
Config_SectionsForbidden_ScalarsForbidden: TypeAlias = Section_ScalarsForbidden
Config_SectionsForbidden_ScalarsAllowed: TypeAlias = Section_ScalarsAllowed
Config_SectionsAllowed_ScalarsAllowed: TypeAlias = (
    dict[str, Section_ScalarsAllowed] | Section_ScalarsAllowed
)
Config_SectionsAllowed_ScalarsForbidden: TypeAlias = (
    dict[str, Section_ScalarsForbidden] | Section_ScalarsForbidden
)

AnyConfig: TypeAlias = (
    Config_SectionsAllowed_ScalarsAllowed
    | Config_SectionsAllowed_ScalarsForbidden
    | Config_SectionsForbidden_ScalarsAllowed
    | Config_SectionsForbidden_ScalarsForbidden
    | Config_SectionsRequired_ScalarsAllowed
    | Config_SectionsRequired_ScalarsForbidden
)



from collections.abc import Iterator
from typing import Any


def always_iterable(obj: Any, /) -> Iterator[Any]:
    # adapted from more_iterools 10.1.0 (MIT)
    if isinstance(obj, str):
        return iter((obj,))

    try:
        return iter(obj)
    except TypeError:
        return iter((obj,))



import sys
from inspect import cleandoc, getdoc
from textwrap import dedent

import pytest

import inifix

if sys.flags.optimize >= 2:  # pragma: no cover
    pytest.skip(
        reason="docstrings are not available in optimized mode 2",
        allow_module_level=True,
    )


PARSING_OPTIONS = """
    parse_scalars_as_lists: bool (default: False)
        if set to True, all values will be parses as lists of scalars,
        even for parameters comprised of a single scalar.
        This also has the effect of narrowing down the return type as seen
        by static type checkers.

        .. versionadded: 3.1.0

    integer_casting: 'stable' (default) or 'aggressive'
        casting strategy for numbers written in decimal notations, such as '1.',
        '2.0' or '3e0'. By default, perform roundtrip-stable casting (i.e., cast
        as Python floats). Setting `integer_casting='aggressive'` will instead
        parse these as Python ints, matching the behavior of inifix 4.5

        .. versionadded: 5.0.0"""

VALIDATION_OPTIONS = """
    sections: 'allow' (default), 'forbid' or 'require'
        use sections='forbid' to invalidate any section found,
        or sections='require' to invalidate a sectionless structure.
        Default mode (sections='allow') permits both.
        sections='forbid' and sections='require' both have the effect of
        narrowing down the return type as seen by static type checkers.
        This parameter has no effect at runtime when combined with
        skip_validation=True.

        .. versionadded: 5.1.0

    skip_validation: bool (default: False)
        if set to True, input is not validated. This can be used to speedup
        io operations trusted input or to work around bugs with the validation
        routine. Use with caution.
        This parameter intentionally has no effect on static type checking.

        .. versionadded: 4.1.0"""

DUMP_DOCSTRING = cleandoc(
    f"""
    Write data to a file.

    Parameters
    ----------
    data: dict
        has to be inifix format-compliant

    file: any of the following
        - the name of a file to write to (str, bytes or os.PathLike)
        - a writable handle. Both text and binary file modes are supported,
          though binary is preferred.
          In binary mode, data is encoded as UTF-8.
    {VALIDATION_OPTIONS}

    See Also
    --------
    inifix.dumps
    """
)

DUMPS_DOCSTRING = cleandoc(
    f"""
    Convert data to a string.

    Parameters
    ----------
    data: dict
        has to be inifix format-compliant
    {VALIDATION_OPTIONS}

    Returns
    -------
    string representing the content of a parameter file

    See Also
    --------
    inifix.dump
    """
)


LOAD_DOCSTRING = cleandoc(
    f"""
    Parse data from a file.

    Parameters
    ----------
    source: any of the following
        - the name of a file to read from, (str, bytes or os.PathLike)
        - a readable handle. Both text and binary file modes are supported,
          though binary is preferred.
          In binary mode, we assume UTF-8 encoding.
    {PARSING_OPTIONS}
    {VALIDATION_OPTIONS}

    See Also
    --------
    inifix.loads
    """
)

LOADS_DOCSTRING = cleandoc(
    f"""
    Parse data from a string.

    Parameters
    ----------
    source: str
        The content of a parameter file (has to be inifix format-compliant)
    {PARSING_OPTIONS}
    {VALIDATION_OPTIONS}

    See Also
    --------
    inifix.load
    """
)


@pytest.mark.parametrize(
    "func, expected_docstring",
    [
        (inifix.load, LOAD_DOCSTRING),
        (inifix.loads, LOADS_DOCSTRING),
        (inifix.dump, DUMP_DOCSTRING),
        (inifix.dumps, DUMPS_DOCSTRING),
    ],
)
def test_docstrings(func, expected_docstring):
    assert getdoc(func) == dedent(expected_docstring)



from importlib.metadata import version

import pytest

import inifix


def test_dunder_version():
    assert inifix.__version__ == version("inifix")


def test_unknown_member():
    with pytest.raises(AttributeError):
        inifix.unknown_member  # noqa: B018



import os
import shutil
from difflib import unified_diff
from pathlib import Path
from stat import S_IREAD

import pytest

import inifix.format
from inifix import load
from inifix.format import format_string, iniformat, main

DATA_DIR = Path(__file__).parent / "data"


@pytest.mark.parametrize("args", [(), ("--skip-validation",)])
def test_format_keep_data(args, inifile, capsys, tmp_path):
    target = tmp_path / inifile.name

    ref_data = load(inifile)
    shutil.copyfile(inifile, target)

    ret = main([str(target), *args])
    assert isinstance(ret, int)

    out, err = capsys.readouterr()

    assert out == ""
    data_new = load(target)
    assert data_new == ref_data


def diff_file(f1: Path, f2: Path) -> str:
    return (
        "\n".join(
            line.removesuffix("\n")
            for line in unified_diff(
                f1.read_text().splitlines(),
                f2.read_text().splitlines(),
                fromfile=str(f1),
                # tofile=str(f2),
            )
        )
        + "\n"
    )


@pytest.mark.parametrize(
    "infile, expect_diff", [("format-in.ini", True), ("format-out.ini", False)]
)
def test_exact_format_diff(infile, expect_diff, capsys):
    body = (DATA_DIR / infile).read_text()

    ret = main([str(DATA_DIR / infile), "--diff"])
    out, err = capsys.readouterr()
    if expect_diff:
        assert ret != 0
        expected = diff_file(DATA_DIR / infile, DATA_DIR / "format-out.ini")
        assert out == expected
        assert err == ""
    else:
        assert ret == 0
        assert out == ""
        assert err == ""

    assert (DATA_DIR / infile).read_text() == body


def test_exact_format_inplace(capsys, tmp_path):
    DATA_DIR = Path(__file__).parent / "data"
    target = tmp_path / "out.ini"
    shutil.copyfile(DATA_DIR / "format-in.ini", target)

    ret = main([str(target)])
    out, err = capsys.readouterr()

    assert err == f"Fixing {target}\n"
    assert ret != 0

    expected = (DATA_DIR / "format-out.ini").read_text()
    res = target.read_text()
    assert res == expected


def test_no_parameters(tmp_path):
    target = tmp_path / "no_params.ini"
    target.write_text(
        "    # comment 1\n"
        "[Section A]\n"
        "  # comment 2\n"
        " # comment 3\n"
        "[Section B]"
    )
    ret = main([str(target)])
    assert ret != 0

    expected = (
        "# comment 1\n"
        "\n"
        "[Section A]\n"
        "# comment 2\n"
        "# comment 3\n"
        "\n"
        "[Section B]\n"
    )
    assert target.read_text() == expected


def test_missing_file(capsys, tmp_path):
    target = tmp_path / "not_a_file"
    ret = main([str(target)])
    assert ret != 0
    out, err = capsys.readouterr()
    assert out == ""
    assert err == f"Error: could not find {target}\n"


def test_empty_file(capsys, tmp_path):
    target = tmp_path / "invalid_file"
    target.touch()
    ret = main([str(target)])
    assert ret != 0
    out, err = capsys.readouterr()
    assert out == ""
    assert f"Error: {str(target)!r} appears to be empty.\n" in err


def test_error_read_only_file(inifile, capsys, tmp_path):
    target = tmp_path / inifile.name
    shutil.copy(inifile, target)

    data = target.read_text()
    if format_string(data) == data:
        return

    os.chmod(target, S_IREAD)

    ret = main([str(target)])
    assert ret != 0
    out, err = capsys.readouterr()
    assert out == ""
    assert f"Error: could not write to {target} (permission denied)\n" in err


def test_informat_depr():
    body = (DATA_DIR / "format-in.ini").read_text()
    s1 = format_string(body)
    with pytest.warns(
        DeprecationWarning,
        match="inifix.format.iniformat is deprecated",
    ):
        s2 = iniformat(body)
    assert s2 == s1


def test_diff_stdout(inifile, capsys, tmp_path):
    target = tmp_path / inifile.name
    shutil.copy(inifile, target)

    ret = main([str(target), "--diff"])
    # we can't predict if formatting is needed
    assert isinstance(ret, int)
    out, err = capsys.readouterr()

    assert (ret == 0 and out == "") or (ret != 0 and err == "" and out != "")


def test_report_noop(capsys, tmp_path):
    inifile = DATA_DIR / "format-out.ini"
    target = tmp_path / inifile.name

    shutil.copyfile(inifile, target)

    ret = main([str(target), "--report-noop"])
    assert ret == 0

    out, err = capsys.readouterr()

    assert out == ""
    assert err == f"{target} is already formatted\n"


def test_format_quoted_strings_with_whitespaces(tmp_path):
    target = tmp_path / "spaces.ini"

    inserted = '''Eggs 'Bacon Saussage'     "spam"'''
    expected = """Eggs    'Bacon Saussage'  "spam"\n"""
    target.write_text(inserted)
    ret = main([str(target)])
    assert ret != 0

    assert target.read_text() == expected


def test_data_preservation(inifile, tmp_path):
    # check that perilous string manipulations do not destroy data
    initial_mapping = load(inifile)
    target = tmp_path / inifile.name
    shutil.copyfile(inifile, target)
    main([str(target)])
    round_mapping = load(target)
    assert round_mapping == initial_mapping


def test_single_core(monkeypatch, tmp_path):
    # regression test for gh-251
    target = tmp_path / "spaces.ini"

    inserted = '''Eggs 'Bacon Saussage'     "spam"'''
    target.write_text(inserted)

    monkeypatch.setattr(inifix.format, "get_cpu_count", lambda: 1)
    inifix.format.main([str(target)])



import re

ENOTATION_REGEXP = re.compile(r"\d+(\.\d*)?e[+-]?\d+?")


class ENotationIO:
    """A small class to encode/decode real numbers to and from
    e-notation formatted strings.

    """

    @staticmethod
    def decode(s: str, /) -> int:
        """
        Cast an 'e' formatted string `s` to integer if such a conversion can
        be performed without loss of data. Raise ValueError otherwise.

        Examples
        --------
        >>> ENotationIO.decode("6.28E2")
        628
        >>> ENotationIO.decode("1.4e3")
        1400
        >>> ENotationIO.decode("7.0000E2")
        700
        >>> ENotationIO.decode("700.00E-2")
        7
        >>> ENotationIO.decode("700e-2")
        7
        >>> ENotationIO.decode("6.000e0")
        6
        >>> ENotationIO.decode("700e-3")
        Traceback (most recent call last):
        ...
        ValueError
        >>> ENotationIO.decode("7.0001E2")
        Traceback (most recent call last):
        ...
        ValueError
        >>> ENotationIO.decode("0.6e0")
        Traceback (most recent call last):
        ...
        ValueError
        >>> ENotationIO.decode("notanumber")
        Traceback (most recent call last):
        ...
        ValueError
        """
        s = s.lower()

        if not re.match(ENOTATION_REGEXP, s):
            raise ValueError

        digits, _, sexponent = s.partition("e")
        exponent = int(sexponent)
        if "." in digits:
            digits, decimals = digits.split(".")
            decimals = decimals.rstrip("0")
        else:
            decimals = ""

        if exponent >= 0 and len(decimals) > exponent:
            raise ValueError
        elif exponent < 0 and len(digits) - 1 < -exponent:
            raise ValueError

        return int(float(s))

    @staticmethod
    def simplify(s: str, /) -> str:
        """
        Simplify exponents and trailing zeros in decimals.
        This is a helper function to `ENotationIO.encode`.

        >>> ENotationIO.simplify('1e-00')
        '1e0'
        >>> ENotationIO.simplify('1e+00')
        '1e0'
        >>> ENotationIO.simplify('1e-01')
        '1e-1'
        >>> ENotationIO.simplify('1e+01')
        '1e1'
        >>> ENotationIO.simplify('1e+10')
        '1e10'
        >>> ENotationIO.simplify('1.000e+00')
        '1e0'
        >>> ENotationIO.simplify('1.100e+00')
        '1.1e0'
        """
        s = re.sub(r"\.?0*(e[+-]?)0", r"\1", s)
        s = re.sub(r"(e-0)$", "e0", s)
        return s.replace("+", "")

    @staticmethod
    def encode(r: float, /) -> str:
        """
        Convert a real number `r` to string, using scientific notation.

        Note that this differs from using format specifiers (e.g. `.6e`)
        in that trailing zeros are removed.
        Precision must be conserved.

        Parameters
        ----------
        r: real number (float or int)

        Returns
        -------
        ret: str
            A string representing a number in sci notation

        Examples
        --------
        >>> ENotationIO.encode(1)
        '1e0'
        >>> ENotationIO.encode(0.0000001)
        '1e-7'
        >>> ENotationIO.encode(10_000_000)
        '1e7'
        >>> ENotationIO.encode(156_000)
        '1.56e5'
        >>> ENotationIO.encode(0.0056)
        '5.6e-3'
        >>> ENotationIO.encode(3.141592653589793)
        '3.141592653589793e0'
        >>> ENotationIO.encode(1e-15)
        '1e-15'
        >>> ENotationIO.encode(0.0)
        '0e0'
        >>> ENotationIO.encode(0)
        '0e0'
        """
        base = str(r)
        if "e" in base:
            return ENotationIO.simplify(base)
        if not base.strip(".0"):
            return "0e0"
        max_ndigit = len(base.replace(".", "")) - 1
        return ENotationIO.simplify(f"{r:.{max_ndigit}e}")

    @staticmethod
    def encode_preferential(r: float, /) -> str:
        """
        Convert a float `r` to string, using sci notation if
        and only if it saves space.

        Examples
        --------
        >>> ENotationIO.encode_preferential(189_000_000)
        '1.89e8'
        >>> ENotationIO.encode_preferential(189)
        '189.0'
        >>> ENotationIO.encode_preferential(900)
        '9e2'
        >>> ENotationIO.encode_preferential(1)
        '1.0'
        >>> ENotationIO.encode_preferential(0.7)
        '0.7'
        >>> ENotationIO.encode_preferential(0.00007)
        '7e-5'
        """
        return min(str(float(r)), ENotationIO.encode(r), key=lambda x: len(x))



import pytest

from inifix.validate import main


@pytest.fixture(
    params=[
        pytest.param("a\n", id="missing value (empty row)"),
        pytest.param("a #\n", id="missing value (comm)"),
        pytest.param(r"/!\\\n", id="invalid param name (special chars)"),
        pytest.param("100 100\n", id="invalid param name (num-1)"),
        pytest.param("1e2 100\n", id="invalid param name (num-2)"),
        pytest.param("[Unclosed Section\na 1 2 3\n", id="unclosed section"),
        pytest.param("[[extra left bracket]\na 1 2 3\n", id="mismatched left bracket"),
        pytest.param(
            "[extra right bracket]]\na 1 2 3\n", id="mismatched right bracket"
        ),
        pytest.param("[extra] stuff that's not a comment", id="missing comment char"),
        pytest.param("[()]", id="invalid section chars (parens)"),
        pytest.param("[{}]", id="invalid section chars (brackets)"),
    ]
)
def invalid_file(tmp_path, request):
    file = tmp_path / "myfile.ini"
    file.write_text(request.param)
    return file


def test_empty_file(capsys, tmp_path):
    target = tmp_path / "invalid_file"
    target.touch()
    ret = main([str(target)])
    assert ret != 0
    out, err = capsys.readouterr()
    assert out == ""
    assert f"Failed to validate {target}:" in err
    assert "appears to be empty.\n" in err


def test_missing_file(capsys, tmp_path):
    target = tmp_path / "not_a_file"
    ret = main([str(target)])
    assert ret != 0
    out, err = capsys.readouterr()
    assert out == ""
    assert f"Error: could not find {target}" in err


def test_invalid_files(invalid_file, capsys):
    ret = main([str(invalid_file)])

    stdout, stderr = capsys.readouterr()

    assert stdout == ""
    assert stderr.startswith("Failed to validate")
    assert ret != 0


def test_valid_files(inifile, capsys):
    ret = main([str(inifile)])

    stdout, stderr = capsys.readouterr()

    assert stdout == f"Validated {inifile}\n"
    assert stderr == ""
    assert ret == 0
